/*
 *
 * nucleo/core/carbon/cFileKeeper.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/core/carbon/cFileKeeper.H>
#include <nucleo/utils/FileUtils.H>

#include <sys/types.h>
#include <sys/socket.h>

#include <iostream>

namespace nucleo {
  
  /*
    Problem #1: The CF callback mechanism doesn't give any clue of the
    current __state__ of the socket (if I receive a
    kCFSocketReadCallBack, how do I know whether it's also writable?).
    
    Solution to problem #1: see FileKeeper::getState

    Problem #2: We sometimes get notified by CF that a TCP socket is
    readable although there is no data available. Moreover,

    Solution to problem #2: This is not really a problem as long as
    ReactiveObjects check the current state of the FileKeeper before
    reacting.

    Problem #3: If the remote peer closes a TCP connection, we do not
    get any notification (while select would say that the socket is
    readable and read would then fail).

    Solution to problem #3: Periodically generate an event for TCP
    client sockets so that ReactiveObjects will check the current
    state...
  */

  // -----------------------------------------------------------------------------

  void
  cFileKeeper::_callback(CFSocketRef s,
					CFSocketCallBackType type,
					CFDataRef address, const void *data, void *info) {
    cFileKeeper *obj = (cFileKeeper *)info ;
    obj->notifyObservers() ;
    CFSocketEnableCallBacks(s, type) ;
  }

  void
  cFileKeeper::_callback2(EventLoopTimerRef timerId, void *data){
    cFileKeeper *obj = (cFileKeeper*)data ;
    // std::cerr << "cFileKeeper::_callback2 " << obj << std::endl ;
    obj->notifyObservers() ;
  }

  void
  cFileKeeper::_cleanup(void) {
    if (_timerId) {
	 RemoveEventLoopTimer(_timerId) ;
	 _timerId = 0 ;
    }
    CFRunLoopRemoveSource(CFRunLoopGetCurrent(), _cfRunLoopSource, kCFRunLoopCommonModes);
    CFSocketInvalidate(_cfSocket);
    CFRelease(_cfRunLoopSource);
    CFRelease(_cfSocket);
    _fd = -1 ;
  }

  void
  cFileKeeper::setup(int fd, eventMask mask) {
    if (_fd!=-1) _cleanup() ;

    FileKeeper::setup(fd, mask) ;

    _flags = kCFSocketNoCallBack ;
    if (mask&FileKeeper::R) _flags = _flags | kCFSocketReadCallBack ;
    if (mask&FileKeeper::W) _flags = _flags | kCFSocketWriteCallBack ;

    CFSocketContext context={0,(void *)this,NULL,NULL,NULL};

    _cfSocket = CFSocketCreateWithNative(NULL, fd, _flags, _callback, &context) ;

    if(!_cfSocket) {
	 fd = -1 ;
	 return ;
    }

    // turn off kCFSocketAutomaticallyReenableRead/WriteCallBack and
    // kCFSocketCloseNativeOnInvalidate
    CFSocketSetSocketFlags(_cfSocket, 0) ;

    _cfRunLoopSource = CFSocketCreateRunLoopSource(NULL, _cfSocket, 0);
    if(!_cfRunLoopSource) {
	 fd = -1 ;
	 CFRelease(_cfSocket) ;
	 return ;
    }

    CFRunLoopAddSource(CFRunLoopGetCurrent(), _cfRunLoopSource, kCFRunLoopDefaultMode);

    int type, size=sizeof (int) ;
    if (!getsockopt(fd, SOL_SOCKET, SO_TYPE, (char *) &type, &size)
	   && type==SOCK_STREAM) {
	 sockaddr hisaddr;
	 socklen_t lenhisaddr = sizeof(sockaddr);
	 if (getpeername(fd, (sockaddr *)&hisaddr, &lenhisaddr)!=-1) {
	   // std::cerr << this << " (fd #" << fd << ") is a SOCK_STREAM socket connected to someone" << std::endl ; 
	   EventTimerInterval eti = 750*kEventDurationMillisecond ;
	   InstallEventLoopTimer(GetMainEventLoop(),
						eti, eti /* repeat */,
						NewEventLoopTimerUPP(_callback2), this, &_timerId) ;
	 }
    }
  }

  cFileKeeper::cFileKeeper(void) {
    _timerId = 0 ;
    // std::cerr << "Creating FileKeeper " << this << std::endl ;
  }

  cFileKeeper::~cFileKeeper(void) {
    // std::cerr << "Deleting FileKeeper " << this << std::endl ;
    if (_fd!=-1) _cleanup() ;
  }

  // -----------------------------------------------------------------------------

}
