/*
 *
 * nucleo/gl/window/glWindow_AGL.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>
#include <nucleo/gl/window/glWindow_AGL.H>
#include <nucleo/core/ReactiveEngine.H>

#include <OpenGL/OpenGL.h>

// For CG*
#include <ApplicationServices/ApplicationServices.h>

#include <stdexcept>

#define DEBUG_LEVEL 0

namespace nucleo {

  static inline void
  debugModifiers(std::string msg, UInt32 modifiers) {
    std::cerr << "kEventClassKeyboard: " << msg << " " << modifiers << std::endl ;
    std::cerr << "   cmdKey: " << (modifiers & cmdKey) << std::endl ;
    std::cerr << "   shiftKey: " << (modifiers & shiftKey) << std::endl ;
    std::cerr << "   alphaLock: " << (modifiers & alphaLock) << std::endl ;
    std::cerr << "   optionKey: " << (modifiers & optionKey) << std::endl ;
    std::cerr << "   controlKey: " << (modifiers & controlKey) << std::endl ;
    std::cerr << "   rightShiftKey: " << (modifiers & rightShiftKey) << std::endl ;
    std::cerr << "   rightOptionKey: " << (modifiers & rightOptionKey) << std::endl ;
    std::cerr << "   rightControlKey: " << (modifiers & rightControlKey) << std::endl ;
    
    std::cerr << "   NumLock: " << (modifiers & kEventKeyModifierNumLockMask) << std::endl ;
    std::cerr << "   Fn: " << (modifiers & kEventKeyModifierFnMask) << std::endl ;
  }

  // -----------------------------------------------------------------------

  glWindow
  *createAGLwindow(long options, long eventmask) {
    return new glWindow_AGL(options, eventmask) ;
  }

  // -----------------------------------------------------------------------

  OSStatus
  glWindow_AGL::_windowEventHandler(EventHandlerCallRef nextHandler,
							 EventRef theEvent,
							 void* userData) {
    glWindow_AGL *obj = (glWindow_AGL*) userData ;

    // -- Standard handler ---

    // If we're fullscreen, standard handlers need not be called...
    if (! (obj->_fsContext)) {
	 OSStatus result = CallNextEventHandler(nextHandler, theEvent) ;
	 if (result!=eventNotHandledErr && result!=noErr) {
	   std::cerr << "glWindow_AGL: next event handler returned " << result << std::endl ;
	   return result ;
	 }
    }

    // -----------------------

    UInt32 eclass=GetEventClass(theEvent),
	 ekind=GetEventKind(theEvent) ;

    bool eventHandled = true ;

    switch(eclass) {

    case kEventClassMouse :
	 switch(ekind){

	 case kEventMouseDown :
	 case kEventMouseUp : {
	   Point point ;
	   GetEventParameter(theEvent, kEventParamMouseLocation, typeQDPoint, NULL,
					 sizeof(Point), NULL, &point) ;
	   if (obj->_fsContext || (FindWindow(point, &(obj->_window)) == inContent)) {
		glWindow::event *e = new glWindow::event ;
		e->type = (ekind==kEventMouseDown) ? glWindow::event::buttonPress : glWindow::event::buttonRelease;
		e->x = point.h ;
		e->y = point.v ;
		if (! obj->_fsContext) {
		  e->x -= obj->_rect.left ;
		  e->y -= obj->_rect.top ;
		}
		EventMouseButton mouseButton ;
		GetEventParameter(theEvent, kEventParamMouseButton, typeMouseButton, NULL, sizeof(EventMouseButton), NULL, &mouseButton) ;
		if (mouseButton!=1)
		  e->button = mouseButton ;
		else {
		  if (obj->_modifiers&controlKey) e->button = 3 ;
		  else if (obj->_modifiers&optionKey) e->button = 2 ;
		  else e->button = 1 ;
		}
		obj->_equeue.push(e) ;
	   }
#if DEBUG_LEVEL>0
	   else
		std::cerr << "kEventMouseDown|kEventMouseUp: " << point.h << "," << point.v << std::endl ;
#endif
	 } break;          

	 case kEventMouseMoved :
	 case kEventMouseDragged : {
	   Point point;
	   GetEventParameter(theEvent, kEventParamMouseLocation, typeQDPoint, NULL, sizeof(Point), NULL, &point);
	   if (obj->_fsContext || (FindWindow(point, &(obj->_window)) == inContent)) {
		glWindow::event *e = new glWindow::event ;
		e->type = glWindow::event::pointerMotion ;
		e->x = point.h ;
		e->y = point.v ;
		if (! obj->_fsContext) {
		  e->x -= obj->_rect.left ;
		  e->y -= obj->_rect.top ;
		}
		obj->_equeue.push(e) ;
	   }
#if DEBUG_LEVEL>0
	   else
		std::cerr << "kEventMouseMoved|kEventMouseDragged: " << point.h << "," << point.v << std::endl ;
#endif
	 } break;

	 case kEventMouseWheelMoved : {
	   EventMouseWheelAxis axis ;
	   GetEventParameter(theEvent, kEventParamMouseWheelAxis, typeMouseWheelAxis, NULL, sizeof(axis), NULL, &axis ) ;

	   SInt32 delta ;
	   GetEventParameter(theEvent, kEventParamMouseWheelDelta, typeSInt32, NULL, sizeof(delta), NULL, &delta ) ;

	   glWindow::event *e = new glWindow::event ;
	   e->type = glWindow::event::wheelMotion ;
	   e->axis = (int)axis ;
	   e->delta = (int)delta ;
	   obj->_equeue.push(e) ;
	 } break ;

	 default :
	   eventHandled = false ;
	   break ;
	 }
	 break;

    case kEventClassKeyboard :
	 switch(ekind){
	 case kEventRawKeyDown :
	 case kEventRawKeyRepeat:
	 case kEventRawKeyUp : {
	   glWindow::event *e = new glWindow::event ;

	   if (ekind==kEventRawKeyDown)
		e->type = glWindow::event::keyPress ;
	   else // Up or Repeat
		e->type = glWindow::event::keyRelease ;

	   char charCode ;
	   GetEventParameter (theEvent, kEventParamKeyMacCharCodes, typeChar, NULL, sizeof(char), NULL, &charCode);

	   UInt32 keyCode ;
	   GetEventParameter(theEvent, kEventParamKeyCode, typeUInt32, NULL, sizeof(UInt32), NULL, &keyCode);

	   bool unknown_key = false ;

	   switch(charCode) {
	   case 16:
		switch(keyCode) {
		case 122: e->keysym = XK_F1 ; e->keystr = "<F1>" ; break ;
		case 120: e->keysym = XK_F2 ; e->keystr = "<F2>" ; break ;
		case 99: e->keysym = XK_F3 ; e->keystr = "<F3>" ; break ;
		case 118: e->keysym = XK_F4 ; e->keystr = "<F4>" ; break ;
		case 96: e->keysym = XK_F5 ; e->keystr = "<F5>" ; break ;
		case 97: e->keysym = XK_F6 ; e->keystr = "<F6>" ; break ;
		case 98: e->keysym = XK_F7 ; e->keystr = "<F7>" ; break ;
		case 100: e->keysym = XK_F8 ; e->keystr = "<F8>" ; break ;
		case 101: e->keysym = XK_F9 ; e->keystr = "<F9>" ; break ;
		case 109: e->keysym = XK_F10 ; e->keystr = "<F10>" ; break ;
		case 103: e->keysym = XK_F11 ; e->keystr = "<F11>" ; break ;
		case 111: e->keysym = XK_F12 ; e->keystr = "<F12>" ; break ;
		default: unknown_key = true ; break ;
		}
		break ;
#define KEYTRANS(cc,sym,str) case cc: e->keysym=sym; e->keystr=str ; break ;
#include <nucleo/gl/window/_macos2keysym.h>
#undef KEYTRANS
	   default:
		unknown_key = true ;
		break;
	   }

	   if (unknown_key) {
		std::cerr << "glWindow_AGL: unknown MacOS key (c=" << (int)charCode << ",k=" << keyCode << ")" << std::endl;
	   } else {
		obj->_equeue.push(e) ;
		if (ekind==kEventRawKeyRepeat) {
		  glWindow::event *e2 = new glWindow::event ;
		  e2->type = glWindow::event::keyPress ;
		  e2->keysym = e->keysym ;
		  e2->keystr = e->keystr ;
		  obj->_equeue.push(e2) ;
		}
	   }
	   
	 } break;

	 case kEventRawKeyModifiersChanged: {
	   UInt32 m ;
	   GetEventParameter(theEvent, kEventParamKeyModifiers,typeUInt32,NULL,sizeof(UInt32),NULL,&m) ;

	   UInt32 mdelta = m^obj->_modifiers ;
	   // debugModifiers("modifiers = ",m) ;
	   obj->_modifiers = m ;

	   if (mdelta & cmdKey) {
		glWindow::event *e = new glWindow::event ;
		e->type = (m&cmdKey) ? event::keyPress : event::keyRelease ;
		e->keysym = XK_Meta_L ;
		e->keystr = "<LeftMeta>" ;
		obj->_equeue.push(e) ;
	   } else if (mdelta & shiftKey) {
		glWindow::event *e = new glWindow::event ;
		e->type = (m&shiftKey) ? event::keyPress : event::keyRelease ;
		e->keysym = XK_Shift_L ;
		e->keystr = "<LeftShift>" ;
		obj->_equeue.push(e) ;
	   } else if (mdelta & rightShiftKey) {
		glWindow::event *e = new glWindow::event ;
		e->type = (m&rightShiftKey) ? event::keyPress : event::keyRelease ;
		e->keysym = XK_Shift_R ;
		e->keystr = "<RightShift>" ;
		obj->_equeue.push(e) ;
	   } else if (mdelta & controlKey) {
		glWindow::event *e = new glWindow::event ;
		e->type = (m&controlKey) ? event::keyPress : event::keyRelease ;
		e->keysym = XK_Control_L ;
		e->keystr = "<LeftControl>" ;
		obj->_equeue.push(e) ;
	   } else if (mdelta & rightControlKey) {
		glWindow::event *e = new glWindow::event ;
		e->type = (m&rightControlKey) ? event::keyPress : event::keyRelease ;
		e->keysym = XK_Control_R ;
		e->keystr = "<RightControl>" ;
		obj->_equeue.push(e) ;
	   } else if (mdelta & optionKey) {
		glWindow::event *e = new glWindow::event ;
		e->type = (m&optionKey) ? event::keyPress : event::keyRelease ;
		e->keysym = XK_Alt_L ;
		e->keystr = "<LeftAlt>" ;
		obj->_equeue.push(e) ;
	   } else if (mdelta & rightOptionKey) {
		glWindow::event *e = new glWindow::event ;
		e->type = (m&rightOptionKey) ? event::keyPress : event::keyRelease ;
		e->keysym = XK_Alt_R ;
		e->keystr = "<RightAlt>" ;
		obj->_equeue.push(e) ;
	   } else if (mdelta & kEventKeyModifierFnMask) {
		glWindow::event *e = new glWindow::event ;
		e->type = (m&kEventKeyModifierFnMask) ? event::keyPress : event::keyRelease ;
		e->keysym = XK_Super_L ;
		e->keystr = "<LeftSuper>" ;
		obj->_equeue.push(e) ;
	   } else if (mdelta & alphaLock) {
		glWindow::event *e = new glWindow::event ;
		e->type = (m&alphaLock) ? event::keyPress : event::keyRelease ;
		e->keysym = XK_Caps_Lock ;
		e->keystr = "<CapsLock>" ;
		obj->_equeue.push(e) ;
	   } else if (mdelta & kEventKeyModifierNumLockMask) {
		glWindow::event *e = new glWindow::event ;
		e->type = (m&kEventKeyModifierNumLockMask) ? event::keyPress : event::keyRelease ;
		e->keysym = XK_Num_Lock ;
		e->keystr = "<NumLock>" ;
		obj->_equeue.push(e) ;
	   }
	 } break ;
	 default :
	   eventHandled = false ;
	   break;
	 }
	 break ;

    case kEventClassWindow :
	 switch(ekind){
	 case kEventWindowClose : {
	   glWindow::event *e = new glWindow::event ;
	   e->type = glWindow::event::destroy;
	   obj->_equeue.push(e) ;
	 } break ;
	 case kEventWindowDrawContent : {
	   glWindow::event *e = new glWindow::event ;
	   e->type = glWindow::event::expose ;
	   obj->_equeue.push(e) ;
	 } break;          
	 case kEventWindowBoundsChanged : {
	   UInt32 attributes = 0 ;
	   GetEventParameter(theEvent, kEventParamAttributes, typeUInt32, 
					 NULL, sizeof(UInt32), NULL, &attributes) ;
	   bool resizing = (attributes & kWindowBoundsChangeSizeChanged) ;

	   GetWindowBounds(obj->_window, kWindowContentRgn, &obj->_rect) ;
	   unsigned int width = obj->_rect.right - obj->_rect.left ;
	   unsigned int height = obj->_rect.bottom - obj->_rect.top;

	   bool doit = true ;

	   if (resizing && obj->_aspect) {
		unsigned int reqWidth = (unsigned int)(height*obj->_aspect) ;
		if (width != reqWidth) {
		  width = reqWidth ;
		  SizeWindow(obj->_window, width, height, false) ;
		  // we did it already thanks to SizeWindow (kind of recursive call...)
		  doit = false ;
		}
	   }

	   if (doit) {
		aglSetCurrentContext(obj->_context) ;
		aglUpdateContext(obj->_context) ;

		glWindow::event *e = new glWindow::event ;
		e->type = glWindow::event::configure ;
		e->x = obj->_rect.left ;
		e->y = obj->_rect.top ;
		e->width = width ;
		e->height = height ;
		obj->_equeue.push(e) ;
	   }
	 } break;

	 default:
	   eventHandled = false ;
	   break ;
	 }
	 break;	 

    default :
	 eventHandled = false ;
	 break ;
    }

    if (eventHandled) {
	 obj->notifyObservers() ;
	 return noErr ;
    }

    char *ptr = (char *)&eclass ;
    std::cerr << "unexpected event: " ;
    std::cerr << "'" << ptr[0] << ptr[1] << ptr[2] << ptr[3] << "' (" << eclass << ")" ;
    std::cerr << " / " << ekind << std::endl ;
    return eventNotHandledErr ;
  }

  AGLPixelFormat
  glWindow_AGL::getPixelFormat(bool fullscreen) {
    int attributeListSize=2 ;
    GLint attributeList[] = { AGL_RGBA,
						AGL_NO_RECOVERY,
						AGL_NONE,
						AGL_NONE,
						AGL_NONE,
						AGL_NONE,
						AGL_NONE,
						AGL_NONE,
						AGL_NONE,
						AGL_NONE,
						AGL_NONE,
						AGL_NONE,
						AGL_NONE,
						AGL_NONE,
						AGL_NONE,
						AGL_NONE,
						AGL_NONE,
    };

    if (fullscreen)
	 attributeList[attributeListSize++] = AGL_FULLSCREEN ;

    if (_options&glWindow::DOUBLE_BUFFER)
	 attributeList[attributeListSize++] = AGL_DOUBLEBUFFER ;

    if (_options&glWindow::DEPTH) {
	 attributeList[attributeListSize++] = AGL_DEPTH_SIZE ;
	 attributeList[attributeListSize++] = 32 ; // was 24
    }

    if (_options&glWindow::STENCIL) {
	 attributeList[attributeListSize++] = AGL_STENCIL_SIZE ;
	 attributeList[attributeListSize++] = 8 ;
    }

    if (_options&glWindow::ALPHA) {
	 attributeList[attributeListSize++] = AGL_ALPHA_SIZE ;
	 attributeList[attributeListSize++] = 8 ;
    }

    GDHandle device = GetMainDevice();
    AGLPixelFormat apf = aglChoosePixelFormat(&device, 1, attributeList) ;

    if (!apf)
	 throw std::runtime_error("glWindow_AGL: no suitable AGL pixel format") ;
      
#if DEBUG_LEVEL>=2
    {
	 GLint id, red, green, blue, depth, alpha, stencil, fullscreen ;
	 aglDescribePixelFormat(apf, AGL_RENDERER_ID, &id) ;
	 aglDescribePixelFormat(apf, AGL_RED_SIZE, &red) ;
	 aglDescribePixelFormat(apf, AGL_GREEN_SIZE, &green) ;
	 aglDescribePixelFormat(apf, AGL_BLUE_SIZE, &blue) ;
	 aglDescribePixelFormat(apf, AGL_ALPHA_SIZE, &alpha) ;
	 aglDescribePixelFormat(apf, AGL_DEPTH_SIZE, &depth) ;
	 aglDescribePixelFormat(apf, AGL_STENCIL_SIZE, &stencil) ;
	 aglDescribePixelFormat(apf, AGL_FULLSCREEN, &fullscreen) ;

	 std::cerr << "glWindow_AGL:"
			 << " id=" << id
			 << " r=" << red
			 << " g=" << green
			 << " b=" << blue
			 << " a=" << alpha
			 << " depth=" << depth
			 << " stencil=" << stencil
			 << " fullscreen=" << fullscreen
			 << std::endl ;

    }
#endif

    return apf ;
  }

  // -----------------------------------------------------------------------

  glWindow_AGL::glWindow_AGL(long options, long eventmask) {
    debugEvents = false ;
    _aspect = 0.0 ;
    _modifiers = 0 ;
    _mapped = false ;

    _options = options ;

    // --------------------

    SetRect(&_rect, 50, 50, 320, 240) ; // SIF, might want to FIXME...
    WindowClass windowClass = kDocumentWindowClass ;
    WindowAttributes attrs = kWindowAsyncDragAttribute | kWindowCloseBoxAttribute | kWindowCollapseBoxAttribute ;
    // kWindowIgnoreClicksAttribute might also be useful someday
    if (_options&FLOATING) {
	 // std::cerr << "FLOATING" << std::endl ;
	 windowClass = kFloatingWindowClass ;
	 attrs = attrs | kWindowSideTitlebarAttribute ;
    } else if (_options&NONRESIZABLE) {
	 // std::cerr << "NONRESIZABLE" << std::endl ;
    } else {
	 // std::cerr << "NORMAL" << std::endl ;
	 attrs = attrs | kWindowFullZoomAttribute | kWindowResizableAttribute ;
    }
    CreateNewWindow(windowClass, attrs, &_rect, &_window) ;			

    // --------------------

    AGLPixelFormat apf = getPixelFormat(false) ;
    _context = _wContext = aglCreateContext(apf, NULL) ;
    _fsContext = 0 ;
    aglDestroyPixelFormat(apf) ;
    aglSetDrawable(_context, GetWindowPort(_window)) ;
    aglSetCurrentContext(_context) ;

    // --------------------

    unsigned int lastEventType = 0;

    _macEvents[lastEventType].eventClass = kEventClassWindow ;
    _macEvents[lastEventType++].eventKind = kEventWindowBoundsChanged ;

    if (eventmask & glWindow::event::pointerMotion){
	 _macEvents[lastEventType].eventClass = kEventClassMouse ;
	 _macEvents[lastEventType++].eventKind = kEventMouseMoved ;
	 _macEvents[lastEventType].eventClass = kEventClassMouse ;
	 _macEvents[lastEventType++].eventKind = kEventMouseDragged ;
    }
    if (eventmask & glWindow::event::wheelMotion){
	 _macEvents[lastEventType].eventClass = kEventClassMouse ;
	 _macEvents[lastEventType++].eventKind = kEventMouseWheelMoved ;
    }
    if (eventmask & glWindow::event::buttonPress){
	 _macEvents[lastEventType].eventClass = kEventClassMouse ;
	 _macEvents[lastEventType++].eventKind = kEventMouseDown ;
    }
    if (eventmask & glWindow::event::buttonRelease){
	 _macEvents[lastEventType].eventClass = kEventClassMouse ;
	 _macEvents[lastEventType++].eventKind = kEventMouseUp ;
    }

    if (eventmask & glWindow::event::keyPress){
	 _macEvents[lastEventType].eventClass = kEventClassKeyboard ;
	 _macEvents[lastEventType++].eventKind = kEventRawKeyDown ;
    }
    if (eventmask & glWindow::event::keyRelease){
	 _macEvents[lastEventType].eventClass = kEventClassKeyboard ;
	 _macEvents[lastEventType++].eventKind = kEventRawKeyUp ;
    }
    if (eventmask & (glWindow::event::keyPress|glWindow::event::keyRelease)){
	 _macEvents[lastEventType].eventClass = kEventClassKeyboard ;
	 _macEvents[lastEventType++].eventKind = kEventRawKeyModifiersChanged ;
	 _macEvents[lastEventType].eventClass = kEventClassKeyboard ;
	 _macEvents[lastEventType++].eventKind = kEventRawKeyRepeat ;
    }

    if (eventmask & glWindow::event::expose){
	 _macEvents[lastEventType].eventClass = kEventClassWindow ;
	 _macEvents[lastEventType++].eventKind = kEventWindowDrawContent ;
    }

    if (eventmask & glWindow::event::destroy) {
	 _macEvents[lastEventType].eventClass = kEventClassWindow ;
	 _macEvents[lastEventType++].eventKind = kEventWindowClose ;
    }

    if (eventmask & glWindow::event::enter)
	 std::cerr << "glWindow_AGL: 'enter' events are not supported..." << std::endl ;
    if (eventmask & glWindow::event::leave)
	 std::cerr << "glWindow_AGL: 'leave' events are not supported..." << std::endl ;

    _nbMacEvents = lastEventType;

    EventTargetRef winTarget = GetWindowEventTarget(_window) ;
    InstallStandardEventHandler(winTarget) ;
    InstallEventHandler(winTarget,
				    NewEventHandlerUPP(glWindow_AGL::_windowEventHandler),
				    _nbMacEvents, _macEvents,
				    (void *)this,
				    &_evHandler) ;

    // So that objects that subscribe to us can get a "configure"
    // event (see the "react" method below)
    ReactiveEngine::notify(this, this) ;
  }
     
  glWindow_AGL::~glWindow_AGL(void) {
    aglSetCurrentContext(NULL) ;
    if (_fsContext) {
	 aglSetDrawable(_fsContext, NULL) ;
	 aglDestroyContext(_fsContext) ;
    }
    aglSetDrawable(_wContext, NULL) ;
    aglDestroyContext(_wContext) ;
    ReleaseWindow(_window) ;
  }

  // -----------------------------------------------------------------------

  void
  glWindow_AGL::setCursorVisible(bool activate) {
    if (activate) ShowCursor() ; else HideCursor() ;
  }

  void
  glWindow_AGL::warpCursor(int x, int y) {
    CGPoint newCursorPosition ;
    newCursorPosition.x = _rect.left + x ;
    newCursorPosition.y = _rect.top + y ;
    CGWarpMouseCursorPosition(newCursorPosition) ;
  }

  void
  glWindow_AGL::syncToVbl(int nb_vbls) {
    GLint n = nb_vbls ;
    aglSetInteger(_context, AGL_SWAP_INTERVAL, &n) ;
  }

  void
  glWindow_AGL::raise(void) {
    SelectWindow(_window) ;
  }

  void
  glWindow_AGL::lower(void) {
    SendBehind(_window, NULL) ;
  }

  float
  glWindow_AGL::getAlpha(void) {
    float alpha=1.0 ;
    GetWindowAlpha(_window, &alpha) ;
    return alpha ;
  }

  void
  glWindow_AGL::setAlpha(float alpha) {
    SetWindowAlpha(_window, alpha) ;
  }

  // -----------------------------------------------------------------------

  unsigned int
  glWindow_AGL::getScreenWidth(void) {
    return (**GetMainDevice()).gdRect.right ;
  }
  
  unsigned int
  glWindow_AGL::getScreenHeight(void) {
    return (**GetMainDevice()).gdRect.bottom ;
  }

  void
  glWindow_AGL::map(void) {
    ShowWindow(_window) ;
    _mapped = true ;
  }
    
  void
  glWindow_AGL::unmap(void) {
    if (!_fsContext) {
	 HideWindow(_window) ;
	 _mapped = false ;
    }
  }

  void
  glWindow_AGL::makeCurrent(void) {
    aglSetCurrentContext(_context) ;      
  }

  void
  glWindow_AGL::swapBuffers(void) {
    aglSwapBuffers(_context) ;
  }

  void
  glWindow_AGL::setTitle(const char *title) {
    Str255 sTitle;
    CopyCStringToPascal(title, sTitle);
    SetWTitle(_window, sTitle);
  }

  void
  glWindow_AGL::getGeometry(unsigned int *width, unsigned int *height, int *x, int *y) {
    if (x) *x = _rect.left ;
    if (y) *y = _rect.top ;
    if (width) *width = _rect.right - _rect.left ;
    if (height) *height = _rect.bottom - _rect.top ;
  }
    
  void 
  glWindow_AGL::setGeometry(unsigned int width, unsigned int height, int x, int y) {
    if (!_mapped) map() ;
    SizeWindow(_window, width, height, false);
    MoveWindow(_window, x, y, false);
  }

  void 
  glWindow_AGL::setGeometry(unsigned int width, unsigned int height) {
    if (!_mapped) map() ;
    SizeWindow(_window, width, height, false);
  }
  void
  glWindow_AGL::setFullScreen(bool activate) {
    if (!activate && !_fsContext) return ;
    if (activate && _fsContext) return ;

    glWindow::event *e = new glWindow::event ;
    e->type = glWindow::event::configure ;

    if (activate) {
	 SetUserFocusWindow(_window) ;
	 CGCaptureAllDisplays() ;

	 AGLPixelFormat apf = getPixelFormat(true) ;
	 _fsContext = aglCreateContext(apf, _wContext) ;
	 aglCopyContext(_wContext, _fsContext, GL_ALL_ATTRIB_BITS) ;
	 GLint swap_interval = 0 ;
	 aglGetInteger(_wContext, AGL_SWAP_INTERVAL, &swap_interval) ;
	 aglSetInteger(_fsContext, AGL_SWAP_INTERVAL, &swap_interval) ;
	 aglSetCurrentContext(_fsContext) ;
	 aglUpdateContext(_fsContext) ;
	 aglSetFullScreen(_fsContext, 0, 0, 0, 0) ;
	 _context = _fsContext ;

	 RemoveEventHandler(_evHandler) ;
	 InstallApplicationEventHandler(NewEventHandlerUPP(glWindow_AGL::_windowEventHandler),
							  _nbMacEvents, _macEvents,
							  (void *)this,
							  &_evHandler) ;

	 GLint params[3] ;
	 aglGetInteger(_fsContext, AGL_FULLSCREEN, params) ;
	 e->width = params[0] ; e->height = params[1] ;
	 e->x = e->y = 0 ;
    } else {
	 aglSetDrawable(_fsContext, NULL) ;
	 aglCopyContext(_fsContext, _wContext, GL_ALL_ATTRIB_BITS) ;
	 aglDestroyContext(_fsContext) ;
	 aglSetCurrentContext(_wContext) ;
	 // aglUpdateContext(_wContext) ;
	 _context = _wContext ;
	 _fsContext = 0 ;

	 CGReleaseAllDisplays() ;
	 ShowWindow(_window) ;
	 _mapped = true ;

	 RemoveEventHandler(_evHandler) ;
	 InstallEventHandler(GetWindowEventTarget(_window),
					 NewEventHandlerUPP(glWindow_AGL::_windowEventHandler),
					 _nbMacEvents, _macEvents,
					 (void *)this,
					 &_evHandler) ;

	 e->width = _rect.right - _rect.left ; e->height = _rect.bottom - _rect.top ;
	 e->x = _rect.left ; e->y = _rect.top ;
    }

    _equeue.push(e) ;
    notifyObservers() ;
  }

  void
  glWindow_AGL::setAspectRatio(int width, int height) {
    _aspect = (double)width/(double)height ;
  }

  // -----------------------------------------------------------------------

  bool
  glWindow_AGL::getNextEvent(event *e) {
    if (_equeue.empty()) return false ;

    event *eq = _equeue.front() ;
    _equeue.pop() ;
    *e = *eq ;
    if (debugEvents) {
	 std::cerr << "glWindow_AGL event: " ;
	 e->debug(std::cerr) ;
	 std::cerr << std::endl ;
    }
    delete eq ;

    return true ;
  }

  // -----------------------------------------------------------------------

  void 
  glWindow_AGL::react(Observable *obs) {
    if (obs==this) {
	 glWindow::event *e = new glWindow::event ;
	 e->type = glWindow::event::configure ;
	 e->width = _rect.right - _rect.left ; e->height = _rect.bottom - _rect.top ;
	 e->x = _rect.left ; e->y = _rect.top ;
	 _equeue.push(e) ;
	 notifyObservers() ;
    }
  }

}
