/*
 *
 * nucleo/image/processing/basic/Paint.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/image/processing/basic/Paint.H>

#include <nucleo/image/encoding/Conversion.H>

namespace nucleo {

  // ------------------------------------------------------------------

  bool
  paintImageRegion(Image *img,
			    unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2,
			    unsigned char R, unsigned char G, unsigned char B, unsigned char A) {
    if (!convertImage(img, Image::CONVENIENT)) return false ;

    unsigned int bytesPerPixel = img->getBytesPerPixel() ;
    unsigned int width = img->getWidth() ;
    unsigned int height = img->getHeight() ;

    if (x1>width-1) x1=width-1 ; if (x2>width-1) x2=width-1 ;
    if (x1<0) x1=0 ; if (x2<0) x2=0 ;

    if (y1>height-1) y1=height-1 ; if (y2>height-1) y2=height-1 ;
    if (y1<0) y1=0 ; if (y2<0) y2=0 ;

    unsigned char *data = (unsigned char *)img->getData() ;
    int lineWidth = width*bytesPerPixel ;

    unsigned char newvalues[4] = {A, R, G, B} ;

    Image::Encoding e = img->getEncoding() ;
    if (e==Image::RGB || e==Image::ARGB) {
	 unsigned int offset = (e==Image::RGB) ? 1 : 0 ;
	 for (unsigned int y=y1; y<=y2; ++y) {
	   unsigned char *ptr = data + lineWidth*y + x1*bytesPerPixel ;
	   for (unsigned int x=x1; x<=x2; ++x) {
		memcpy(ptr, newvalues+offset, bytesPerPixel) ;
		ptr += bytesPerPixel ;
	   }
	 }
    } else if (e==Image::L) {
	 for (unsigned int y=y1; y<=y2; ++y) {
	   unsigned char *ptr = data + lineWidth*y + x1 -1 ;
	   for (unsigned int x=x1; x<=x2; ++x)
		*ptr++ = (int)(0.11*B + 0.59*G + 0.3*R) ;
	 }
    } else if (e==Image::A) {
	 for (unsigned int y=y1; y<=y2; ++y) {
	   unsigned char *ptr = data + lineWidth*y + x1 -1 ;
	   for (unsigned int x=x1; x<=x2; ++x)
		*ptr++ = A ;
	 }
    }

    return true ;
  }

  // ------------------------------------------------------------------

  bool
  paintImage(Image *img,
		   unsigned char R, unsigned char G, unsigned char B, unsigned char A) {
    unsigned int width = img->getWidth() ;
    unsigned int height = img->getHeight() ;
    if (!width || !height) return false ;

    return paintImageRegion(img,
					   0,0,width-1,height-1,
					   R,G,B,A) ;
  }

  // ------------------------------------------------------------------

  static void
  place(unsigned char *src, const unsigned sWidth, const unsigned sHeight,
	   unsigned char *dst, const unsigned dWidth, const unsigned dHeight,
	   int xD, int yD,
	   const unsigned int bytesPerPixel) {
    if (!src || !sWidth || !sHeight || !dst || !dWidth || !dHeight) return ;

    int xS=0, yS=0 ;
    int w = sWidth ;
    int h = sHeight ;

    if (xD<0) { xS-=xD ; w+=xD ; xD=0 ; }
    if (yD<0) { yS-=yD ; h+=yD ; yD=0 ; }

    if ((unsigned int)xD>=dWidth || (unsigned int)yD>=dHeight) return ;

    if ((unsigned int)(xD+w)>=dWidth) { w=dWidth-xD ; }
    if ((unsigned int)(yD+h)>=dHeight) { h=dHeight-yD ; }

#if 0
    std::cerr << "place: " << xS << "," << yS << " (" << sWidth << "x" << sHeight << ")" ;
    std::cerr << " -- " << w << "x" << h << " --> " ;
    std::cerr << xD << "," << yD << " (" << dWidth << "x" << dHeight << ")" << std::endl ;
#endif
  
    unsigned char *pSrc = (unsigned char *)src + (yS*sWidth + xS)*bytesPerPixel ;
    unsigned char *pDst = dst + (yD*dWidth + xD)*bytesPerPixel ;
    const unsigned int realSWidth = sWidth*bytesPerPixel ;
    const unsigned int realDWidth = dWidth*bytesPerPixel ;
    const unsigned int realWidth = w*bytesPerPixel ;

    while (h--) {
	 memcpy(pDst, pSrc, realWidth) ;
	 pSrc+=realSWidth ;
	 pDst+=realDWidth ;
    }
  }

  // ------------------------------------------------------------------

  bool
  drawImageInImage(Image *src, Image *dst, int x, int y) {
    if (!dst->getData() || !dst->getSize()) return false ;

    if (!convertImage(dst, Image::CONVENIENT)) return false ;

    unsigned int bytesPerPixel = dst->getBytesPerPixel() ;

    Image tmp(*src) ;

    Image::Encoding encoding = dst->getEncoding() ;
    if (tmp.getEncoding() != encoding)
	 if (!convertImage(&tmp, encoding)) return false ;

    place((unsigned char *)tmp.getData(), tmp.getWidth(), tmp.getHeight(),
		(unsigned char *)dst->getData(), dst->getWidth(), dst->getHeight(),
		x, y,
		bytesPerPixel) ;

    return true ;
  }

  // ------------------------------------------------------------------

  bool
  blendImages(Image *src1, Image *src2, Image *dst, float blending) {
    Image tmp1(*src1), tmp2(*src2) ;

    unsigned int wsrc1=tmp1.getWidth(), hsrc1=tmp1.getHeight() ;
    unsigned int wsrc2=tmp2.getWidth(), hsrc2=tmp2.getHeight() ;
    if (wsrc1!=wsrc2 || hsrc1!=hsrc2) return false ;

    Image::Encoding esrc1 = tmp1.getEncoding() ;
    Image::Encoding esrc2 = tmp2.getEncoding() ;
    
    Image::Encoding encoding = esrc1 ;
    if (esrc1!=esrc2 || !tmp1.encodingIsConvenient()) {
	 encoding = Image::ARGB ;
	 convertImage(&tmp1, encoding) ;
	 convertImage(&tmp2, encoding) ;
    }

    unsigned int bytesPerPixel = tmp1.getBytesPerPixel() ;
    unsigned int size = wsrc1*hsrc1*bytesPerPixel ;
    unsigned char *data = Image::AllocMem(size) ;

    float nb = 1.0-blending ;

    unsigned char *ptr1=tmp1.getData()-1, *ptr2=tmp2.getData()-1, *ptr=data-1 ;
    for (int i=size; i; --i) {
	 unsigned char v = (unsigned char)(nb*(*(++ptr1))+blending*(*(++ptr2))) ;
	 *(++ptr) = v ;
    }
    
    dst->setEncoding(encoding) ;
    dst->setDims(wsrc1, hsrc1) ;
    dst->setData(data, size, Image::FREEMEM) ;

    return true ;
  }

  // ------------------------------------------------------------------

  bool
  blendImages(Image *src1, Image *src2, Image *dst) {
    unsigned int wsrc1=src1->getWidth(), hsrc1=src1->getHeight() ;
    unsigned int wsrc2=src2->getWidth(), hsrc2=src2->getHeight() ;
    if (wsrc1!=wsrc2 || hsrc1!=hsrc2) return false ;   

    Image::Encoding esrc2 = src2->getEncoding() ;
    if (esrc2!=Image::ARGB) return false ;

    Image tmp1(*src1) ;
    Image::Encoding esrc1 = tmp1.getEncoding() ;
    if (esrc1!=Image::ARGB && esrc1!=Image::RGB && esrc1!=Image::L) {
	 esrc1 = Image::ARGB ;
	 if (!convertImage(&tmp1, esrc1)) return false ;
    }

    unsigned int bytesPerPixel = tmp1.getBytesPerPixel() ;
    unsigned int nbpixels = wsrc1*hsrc1 ;
    unsigned int size = nbpixels*bytesPerPixel ;
    unsigned char *data = Image::AllocMem(size) ;

    unsigned char *ptr1=tmp1.getData(), *ptr2=src2->getData(), *ptr=data ;

    if (esrc1==Image::L) {
	 for (unsigned int i=0; i<nbpixels; ++i) {
	   float alpha = ptr2[0]/255.0 ;
	   float omalpha = 1.0 - alpha ;
	   float l = omalpha*(*ptr1) + alpha*(0.11*ptr2[3] + 0.59*ptr2[2] + 0.3*ptr2[1]) ;
	   *ptr = (unsigned char)l ;
	   ptr1 ++ ;
	   ptr2 += 4 ;
	   ptr ++ ;
	 }
    } else if (esrc1==Image::RGB) {
	 for (unsigned int i=0; i<nbpixels; ++i) {
	   float alpha = ptr2[0]/255.0 ;
	   float omalpha = 1.0 - alpha ;
	   for (unsigned int c=0; c<3; ++c)
		ptr[c] = (unsigned char)(omalpha*ptr1[c]+alpha*ptr2[c+1]) ;
	   ptr1 += 3 ;
	   ptr2 += 4 ;
	   ptr += 3 ;
	 }
    } else {
	 for (unsigned int i=0; i<nbpixels; ++i) {
	   float alpha = ptr2[0]/255.0 ;
	   float omalpha = 1.0 - alpha ;
	   ptr[0] = ptr1[0] ;
	   for (unsigned int c=1; c<4; ++c)
		ptr[c] = (unsigned char)(omalpha*ptr1[c]+alpha*ptr2[c+1]) ;
	   ptr1 += 4 ;
	   ptr2 += 4 ;
	   ptr += 4 ;
	 }
    }

    dst->setEncoding(esrc1) ;
    dst->setDims(wsrc1, hsrc1) ;
    dst->setData(data, size, Image::FREEMEM) ;  

    return 1 ;
  }

  // ------------------------------------------------------------------
  // Below is some unused old stuff from videoSpace...
  // ------------------------------------------------------------------

#if 0
  int
  setTransparencyLevels(Image *img,
				    unsigned char R, unsigned char G, unsigned char B,
				    unsigned char A) {
    Image::Encoding e = img->getEncoding() ;
    int nbpixels = img->getWidth()*img->getHeight() ;
    unsigned char *ptr = (unsigned char *)img->getData() ;
    int i ;

    switch (e) {
    case Image::ABGR:
	 for (i=0; i<nbpixels; ++i) {
	   *ptr = (*(ptr+1)==B && *(ptr+2)==G && *(ptr+3)==R) ? 0 : A ;  
	   ptr+=4 ;
	 }
	 break ;
    case Image::RGBA:
	 for (i=0; i<nbpixels; ++i) {
	   *(ptr+3) = (*(ptr+2)==B && *(ptr+1)==G && *ptr==R) ? 0 : A ;  
	   ptr+=4 ;
	 }
	 break ;
    default:
	 std::cerr << "setTransparencyLevels works only on RGBA or ABGR images..." << std::endl ;
	 return 0 ;
    }
    return 1 ;
  }
#endif

  // ------------------------------------------------------------------

}
